(************************************NOTICE***********************************
 * This source code and its compiled binary code are  Copyright (C) 1998  by *
 * The Eclectic Company, 1911 Packard, Ann Arbor, MI 48104-4733.  License is *
 * hereby granted to copy,  distribute,  and use them freely for non-commer- *
 * cial research purposes only,  provided:  (1) that this notice is included *
 * in the source code, and (2) that the "trademark" procedure is included in *
 * the compiled binary code.  Modifications and  recompilations of this code *
 * may also be performed freely for the purposes of non-commercial research, *
 * under the same conditions;  the author would  appreciate receiving a copy *
 * of any such modifications.                                                *
 *****************************************************************************)

{
     Program MONOSYL - A data-base manipulation program for monosyllables.

     Written by: John M. Lawler
                 Program in Linguistics
                 University of Michigan
                 Ann Arbor, MI 48109-1285

     Version 1.0  compiled in TurboPascal for CP/M-80.  5/03/84
     Version 5.71 compiled in TurboPascal for DOS       8/10/95
     Version 6.0  compiled in TurboPascal for DOS       5/4/98

     Use INDEX.COM (source INDEX.PAS) for management.

     Still to do:

          A) Menu system
               1) top/bottom prompt lines  la LIST.COM
               2) Boolean choice interface (language?)
               3) display of lists to choose from
               4) improved Function Key access

          B) Boolean access
               1) OR for subcategories (unions)
                     a) vowel qualities (primarily in rimes)
                        (1) high/low
                        (2) front/back
                        (3) lax/tense
                        (4) mono/diphthong
                     b) cluster types
                        (1) ZERO
                        (2) single cons
                        (3) double 
                            (a) increasing sonority (e.g, FL-, -KS)
                            (b) decreasing sonority (e.g, ST-, -LK)
                        (5) triple, etc.
                        (6) position
                            (a) initial (in assonances)
                            (b) final (in rimes & suffixes)
                     c) consonant features
               2) AND for intersections (e.g, FL + @L)
               3) NOT for complements ("except")
               4) Parentheses for grouping

          C) Character improvement
               1) lower-case (done in 6.0)
               2) phonemic characters:
                     a) vowels
                         (1) ash    (currently AE)
                         (2) open o (currently O; tense /o/ is OW) 
                         (3) open e (currently E; tense /e/ is EY)
                         (4) optional lax high vowels
                            (currently use Trager-Smith I/IY, U/UW)
                     b) consonants
                         (1) theta  (currently TH)
                         (2) edh    (currently DH)
                         (3) angma  (currently N before G or K)
                         (4) hacek s, z, c, j
                              (currently $, ZH, CH, J, respectively)
               3) display, entry, & printing
               4) filter for MONOSYL.DAT

          D) Coding of semantic indices (currently using category numbers
             from Roget's International Thesaurus [4th Edition, ed. R.
             Chapman, Harper & Row 1977]) and extraction of statistical data.
}

type status = (on, off);

const     {-- miscellaneous constants --}
     lower      : set of char = ['a'..'z'];
     upper      : set of char = ['A'..'Z'];
     blank      = ' ';
          {-- useful for filename array --}
     ndx        : status = on;
     ndi        : status = off;
          {-- file names --}
     dataname   = 'MONOSYL.DAT'; {data file}
     signame    = 'MONOSYL.INI'; {initialization file}
          {-- used to check if new trademark needed --}
     trademarked: boolean = false;
          {-- lengths of various strings in database --}
     spellength = 12;  
     asslength  = 4;
     rimlength  = 6;
     suflength  = 3;
     wordlength = 25;
          {-- used for vowel part only; not implemented yet --}
     suffsize   = 20;
     assize     = 75;
             {-- increase the following only if necessary --}
     filesize = 5500; {maximum number of words in database}
     rimesize = 490;  {maximum number of rimes}

type whatsit    = (rim, ass, suf, rand);
     screenline = string [80];
     digraph    = string [4];
     cr         = array [1..2] of integer;
     word       = string [wordlength];
     spelling   = string [spellength];
     assonance  = string [asslength];
     rime       = string [rimlength];
     suffix     = string [suflength];
     rime_ptr   =^rime;

     list_ptr   =^list;
     list       = record
           no   : integer;
           next : list_ptr;
             end; {record list definition}

     index_list = record
          num   : integer;
          name  : rime;
          list  : list_ptr
             end; {record index_list definition}

     entry_ptr  =^entry;
     entry      = record
       spelling : spelling;
       assonance: assonance;
       rime     : rime;
       suffix   : suffix;
             end; {record entry definition}

     regrec     = record
          ax, bx, cx, dx, bp, si, di, ds, es, flags : integer
             end; {record regrec definition}
     date       = record
          year, month, day, hour, minute, second : byte
             end; {record date definition}

     index_type = array [whatsit, 1..filesize] of integer;
     file_type  = array [whatsit, status] of screenline;

var  i,
     j, 
     k,
     offset, 
     number    : integer;
     heaptop   :^integer;
     c, 
     cc        : char;
     escape,
     found, 
     done,
     firstrand : boolean;
     sigfile,
     thisfile,
     datafile  : text;
     testform  : rime;
     sq, 
     temporary : word;
     example, 
     pron      : string [10];
     location,
     topline   : screenline;
     adrs, 
     count     : array [whatsit, 1..rimesize] of integer;
     the_list  : array [whatsit, 1..rimesize] of rime_ptr;
     ndx_file  : array [whatsit] of text;
     numof     : array [whatsit] of integer;
     ndi_file  : array [whatsit] of file of integer;
     ramfile   : array [1..filesize] of entry_ptr;
     address   : array [1..filesize] of integer;
     itsa, 
     lastone   : whatsit;
     indx      :^index_type;
     lastloc   : cr;
     date_1    : date;
     filename  : file_type;

(*
     {this next part is still in development, though it all works...
      BEGIN definitions for vowels, not implemented yet:}

type
     kind_of_vowel = (tense_I, lax_I, tense_E, open_E, ash,
                      SCHWA,
                      A, open_O, tense_O, lax_U, tense_U,
                      AY, AW, OY,
                      error);
     height        = (high, mid, low);
     position      = (front, central, back);
     tenseness     = (tense, neither, lax);
     phthonginess  = (monophthong, diphthong);

     vocoid = record
          height : height;
          position : position;
          tenseness : tenseness;
          phthonginess : phthonginess;
          offset_direction : position;
       end; {record vocoid}

const
     vowel_name : array [kind_of_vowel] of digraph =
                ('IY', 'I', 'EY', 'E', 'AE', '@', 'A',
                 'O', 'OW', 'U', 'UW', 'AY', 'AW', 'OY',
                 'NONE');

     vowel : array [kind_of_vowel] of vocoid =
           ((  {iy}
               height : high;
               position : front;
               tenseness : tense;
               phthonginess : monophthong;
               offset_direction : front),
            (  {i}
               height : high;
               position : front;
               tenseness : lax;
               phthonginess : monophthong;
               offset_direction : central),
            (  {ey}
               height : mid;
               position : front;
               tenseness : tense;
               phthonginess : monophthong;
               offset_direction : front),
            (  {e}
               height : mid;
               position : front;
               tenseness : lax;
               phthonginess : monophthong;
               offset_direction : central),
            (  {ae}
               height : low;
               position : front;
               tenseness : neither;
               phthonginess : monophthong;
               offset_direction : central),
            (  {@}
               height : mid;
               position : central;
               tenseness : neither;
               phthonginess : monophthong;
               offset_direction : central),
            (  {a}
               height : low;
               position : central;
               tenseness : neither;
               phthonginess : monophthong;
               offset_direction : central),
            (  {o}
               height : mid;
               position : back;
               tenseness : lax;
               phthonginess : monophthong;
               offset_direction : central),
            (  {ow}
               height : mid;
               position : back;
               tenseness : tense;
               phthonginess : monophthong;
               offset_direction : back),
            (  {u}
               height : high;
               position : back;
               tenseness : lax;
               phthonginess : monophthong;
               offset_direction : central),
            (  {uw}
               height : high;
               position : back;
               tenseness : tense;
               phthonginess : monophthong;
               offset_direction : back),
            (  {ay}
               height : low;
               position : central;
               tenseness : neither;
               phthonginess : diphthong;
               offset_direction : front),
            (  {aw}
               height : low;
               position : central;
               tenseness : neither;
               phthonginess : diphthong;
               offset_direction : back),
            (  {oy}
               height : mid;
               position : back;
               tenseness : neither;
               phthonginess : diphthong;
               offset_direction : front),
            (  {error}                       {nonsense parameters}
               height : high;
               position : central;
               tenseness : neither;
               phthonginess : diphthong;
               offset_direction : central));

var  rimevowelist   : array [1..rimesize] of kind_of_vowel;
     sufvowelist    : array [1..suffsize] of kind_of_vowel;

function getvowel (thisword : word) : kind_of_vowel;
     {tested and works OK -- not implemented yet in MONOSYL}
const semivowels : set of char  = ['W', 'E', 'Y', 'O'];
var   thisvowel  : digraph;
      ccr        : char;
      thisone    : kind_of_vowel;
      i          : integer;
      empty      : boolean;

begin
    thisvowel := 'NONE';          {set error case as}
    getvowel  := error;           {default value}
    empty := true;
    for i := 1 to length (thisword) do   {eliminate empty case}
        if thisword[i] <> ' ' then       {..check for non-blanks}
            empty := false;
    if not empty then
     begin
      if length (thisword) > 0 then
         thisvowel := copy (thisword, 1, 2);
      if not (thisvowel[2] in semivowels) then
        delete (thisvowel, 2, 1);  {isolate the vowels}
      for thisone := tense_i to oy do
       if thisvowel = vowel_name[thisone] then
         begin                  {check to see if it matches any of the vowels}
           getvowel := thisone; {since this can only happen once,}
           exit;                {.. return when it happens}
         end; {for and if}
      end; {if not empty}
end;  {kind_of_vowel function getvowel}

{END vowel definitions.  Not implemented (or debugged) yet in MONOSYL,
 though everything has been tested and compiles OK}
*)

{ everything below is part of the regular MONOSYL source code:
}

{BEGIN screen and keyboard routines}

function downcase (c:char): char;
begin
     if c in upper then
       downcase := char(ord(c)+32)
     else
       downcase := c;
end;

function getkey : char;
var ch: char;
begin
     escape := false;
     read (kbd, ch);
     if   (ch = #27)
      and keypressed then
         begin
            read (kbd, ch);
            escape := true;
         end;
     getkey := ch;
end;

procedure reversv (text:screenline);
begin
     textbackground (white);
     textcolor (black);
     write (text);
     textbackground (black);
     textcolor (white);
end; {procedure reversv}

procedure queryloc;
begin
     lastloc[1] := wherex;
     lastloc[2] := wherey;
end; {procedure queryloc}

procedure at (loc : cr);
begin
     gotoxy (loc[1], loc[2]);
end; {procedure at}

procedure trademark;
begin
     trademarked := true;
     clrscr; clrscr;
     write ('     ');
     reversv ('              Monosyllable DataBase  *  Version 6.0               ');
     writeln;
     write ('     ');
     reversv ('             Copyright (C) 1998  The Eclectic Company             ');
     writeln;
end; {procedure trademark}

{END screen and keyboard routines;
 --------------------------------
 BEGIN file loading routines}

procedure parse (x : word; var y : entry);
     {converts line entry into record}
var index : integer;
begin
     index := 1;                             {start at char #1}
     while length (x) < wordlength do        {pad x out to full length}
          x := x + ' ';
     with y do      {fill the record structure...}
       begin
         spelling  := copy (x, index, spellength); {strip the spelling}
         index := index + spellength; {advance the counter}
         assonance := copy (x, index, asslength);  {strip the assonance}
         index := index + asslength; {advance the counter}
         rime      := copy (x, index, rimlength);  {strip the rime}
         index := index + rimlength; {advance the counter}
         suffix    := copy (x, index, suflength);  {strip the suffix}
       end; {with statement}
end; {procedure parse}

 procedure getmainin;
 var linotype : word;
 begin
{$I-}
     reset (datafile);
{$I+}
     if ioresult <> 0 then  {this should have been checked already!}
        begin
          gotoxy (10,6);
          clreol;
          writeln ('Data file (' + dataname + ') not found.',
                   '  Terminating.');
          gotoxy (10,7);
          reversv (' Can''t get here from there! ');
          halt;
       end;
     number := 0;
     while not eof (datafile) do  {read in the entries, 1 to a line}
       begin
          number := number + 1;   {the number of words in the db}
          new (ramfile[number]);
          readln (datafile, linotype);       {read it in}
          parse (linotype, ramfile[number]^); {stuff it into memory}
          if number mod 97 = 0 then          {give some feedback}
             begin
               at (lastloc);
               write ('Word #', number);
             end;
       end; {while}
     close (datafile);
     at (lastloc);
     clreol;
     write (number, ' Words');
     write ('.');
     mark (heaptop);
 end; {procedure getmainin}

procedure check_indices;
{
 main routine to check indices for the database.
 uses several low-level routines to compare date stamps and
   check for  file integrity.
 currently does not do updating as such,
     but rather gives message to run INDEX.COM.
}

var handle      : integer;
    file_status : byte;
const read_only = 2;

 procedure update;       {-internal to check_indices}
 var message : screenline;
 begin
     writeln;
     message :=
'  Index Files need updating.  You must run INDEX.COM to recreate them.  ';
     write ('  ');
     reversv (message);
     writeln;
     message :=
'      Then you can run this program again with the updated indices.     ';
     write ('  ');
     reversv (message);
     writeln;
     writeln;
     halt;
 end; {procedure update -internal to check_indices}

 procedure open_file (    filename : screenline;  {-internal to check_indices}
                          mode     : byte;
                      var f_handle : integer;
                      var f_status : byte);
 var regs: regrec;
 begin
     with regs do
       begin
          filename := filename + chr(0);
          ax := $3D00 + mode;
          ds := seg (filename[1]);
          dx := ofs (filename[1]);
          msdos (regs);
          if (1 and flags) = 1 then     {carry set}
               f_status := lo (ax)
          else
            begin
               f_handle := ax;
               f_status := 0            {success}
            end;
       end; {with}
 end; {procedure open_file -internal to check_indices}

 procedure file_stamp (    f_handle : integer; {-internal to check_indices}
                       var d_date   : date;
                       var f_status : byte);
 var regs: regrec;
 begin
     with regs do
        begin
          ax := $5700;
          bx := f_handle;
          msdos (regs);
          if (1 and flags) = 1 then          {carry set}
              f_status := lo (ax)
          else
            begin
              f_status := 0;                 {success}
              with d_date do
                 begin
                    year   := (hi (dx) shr 1) + 1980;
                    month  := (dx and $1E0) shr 5;
                    day    := dx and $1F;
                    hour   := hi (cx) shr 3;
                    minute := (cx and $7E0) shr 5;
                    second := (cx and $1F) shl 1
                 end; {with d_date}
            end; {else}
        end; {with regs}
 end; {procedure file_stamp -internal to check_indices}

 function bad_index (name : screenline) : boolean;
     {-internal to check_indices}
 var  d_x         : date;
 begin
      open_file (name, read_only, handle, file_status);
      if file_status <> 0 then
        begin
          bad_index := true;
          exit;
        end;
      file_stamp (handle, d_x, file_status);
      if file_status <> 0 then
        begin
          bad_index := true;
          exit;
        end;
      with date_1 do
       if year   <> d_x.year   then bad_index := year   > d_x.year   else
       if month  <> d_x.month  then bad_index := month  > d_x.month  else
       if day    <> d_x.day    then bad_index := day    > d_x.day    else
       if hour   <> d_x.hour   then bad_index := hour   > d_x.hour   else
       if minute <> d_x.minute then bad_index := minute > d_x.minute else
                                    bad_index := second > d_x.second;
 end; {boolean function bad_index -internal to check_indices}

 function get_init : screenline;
 var temp_string : screenline;
     leng        : integer;
 begin
      getdir (0, temp_string);      {get the default directory}
      assign (sigfile, signame);    {look for initialization file}
{$I-}
      reset (sigfile);
{$I+}
      if ioresult = 0 then          {if found...}
         begin
            readln (sigfile, temp_string);    {get the directory}
            close (sigfile);
         end
      else
         begin
           assign (sigfile, '\'+signame);     {if not, look in root directory}
{$I-}
           reset (sigfile);
{$I+}
           if ioresult = 0 then               {if it's there...}
              begin
                 readln (sigfile, temp_string);  {get the directory}
                 close (sigfile);
              end; {if}
         end; {else}                      
      leng := length (temp_string);
      if not (temp_string[leng] in ['\', ':']) then
         temp_string := temp_string + '\';
      get_init := temp_string;           {return directory}
 end; {screenline function get_init}

var           ok : boolean;
            leng : integer;          
               cc,
     query_string: screenline;

begin      {procedure check_indices proper}
     trademark;
     filename[ass][ndx]  := 'ASSON.NDX';
     filename[rim][ndi]  := 'RIME.NDI';
     filename[rim][ndx]  := 'RIME.NDX';
     filename[suf][ndi]  := 'SUFFIX.NDI';
     filename[suf][ndx]  := 'SUFFIX.NDX';
     done := false;
     gotoxy (18,4);
     queryloc;
     location := get_init;
     ok := false;
     while not ok do
       begin
          open_file (location+dataname,
                     read_only, handle, file_status);
          if file_status <> 0 then
             begin
               query_string :=
     ' What drive is '+dataname+' on? [default = this one]   '^H^H^H;
               gotoxy (6,5);
               reversv (query_string);
               read (kbd, c);
               if c = ^M then
                begin
                   getdir (0, cc);
                   c := cc[1];
                end;
               c := upcase (c);
               if c = 'Q' then halt;
               query_string :=
     ' What directory is '+dataname+' in?   ("Q" to quit) ';
               gotoxy (6,7);
               reversv (query_string);
               query_string :=  ' ' + c +
     ':                                                ';
               gotoxy (6,8);
               reversv (query_string);
               gotoxy (9,8);
               readln (cc);
               leng := length (cc);
               for i := 1 to leng do
                  cc[i] := upcase (cc[i]);
               if cc = 'Q' then halt;
               if cc[leng] = '\' then
                  delete (cc, leng, 1);
               location := c + ':' + cc;
               trademark;
               assign (sigfile, signame);
               rewrite (sigfile);
               writeln (sigfile, location);
               close (sigfile);
             end {if}
          else ok := true;
       end; {while}
     if location [length(location)] <> '\' then
        location := location + '\';
     file_stamp (handle, date_1, file_status);
     if file_status <> 0 then
        begin
          writeln;
          writeln ('Data File (', dataname, ') damaged.  ',
                 'Terminating.');
          halt;
        end;
     if bad_index (location+filename[ass][ndx]) then update else
     if bad_index (location+filename[rim][ndi]) then update else
     if bad_index (location+filename[rim][ndx]) then update else
     if bad_index (location+filename[suf][ndi]) then update else
     if bad_index (location+filename[suf][ndx]) then update;
end; {procedure check_indices}

procedure startup;
{
 main procedure to load files
}

 procedure get_in (x: whatsit); {- internal to procedure startup}
 begin
     if x = rand then
       begin
          writeln ('get_in() called with "rand" as argument.  Terminating.');
          reversv (' Can''t get here from there! ');
          halt;
       end;
     if x <> ass then
      begin
{$I-}
       reset (ndi_file[x]);
{$I+}
       if ioresult <> 0 then
        begin
          gotoxy (10,6);
          clreol;
          writeln ('Index file (' + filename[x][ndi] + ') not found.',
                   '  Terminating.');
          reversv (' Can''t get here from there! ');
          halt;
        end;
       i := 0;
       while not eof (ndi_file[x]) do
        begin
          i := i + 1;
          read (ndi_file[x], indx^[x,i]);
        end;
       close (ndi_file[x]);
       write ('.');
      end; {if it's not an assonance...}
{$I-}
     reset (ndx_file[x]);
{$I+}
     if ioresult <> 0 then
        begin
          gotoxy (10,6);
          clreol;
          writeln ('Index file (' + filename[x][ndx] + ') not found.',
                   '  Terminating.');
          reversv (' Can''t get here from there! ');
          halt;
        end;
     numof[x]:= 0;
     if x = ass then adrs[x][1] := 0;
     while not eof (ndx_file[x]) do
       begin
          numof[x]:= numof[x] + 1;
          new (the_list[x][numof[x]]);
          read (ndx_file[x], the_list[x][numof[x]]^,
                count[x][numof[x]]);
          if x = ass then adrs[x][numof[x]+1] :=
                          adrs[x][numof[x]]   +  count[x][numof[x]]
          else read (ndx_file[x], adrs[x][numof[x]]);
          readln (ndx_file[x]);
(*  {BEGIN more vowel stuff...}
          case x of
           rim : rimevowelist[numof[rim]] := 
                 getvowel (the_list[rim][numof[rim]]^);
           suf : if the_list[suf][numof[suf]]^ = '   ' then
                    sufvowelist[numof[suf]] := error   else
                 if the_list[suf][numof[suf]]^ = 'LI ' then
                    sufvowelist[numof[suf]] := lax_i   else
                    sufvowelist[numof[suf]] :=
                    getvowel (the_list[suf][numof[suf]]^);
            end; {case x}
*)  {END vowel stuff...}
       end; {while}
     close (ndx_file[x]);
     write ('.');
 end; {procedure get_in - internal to procedure startup}

begin {procedure startup proper}
     at (lastloc);
     clreol;
     write ('Reading input files: ');
     queryloc;
     assign (datafile,      location+dataname);
     assign (ndx_file[ass], location+filename[ass][ndx]);
     assign (ndi_file[rim], location+filename[rim][ndi]);
     assign (ndx_file[rim], location+filename[rim][ndx]);
     assign (ndi_file[suf], location+filename[suf][ndi]);
     assign (ndx_file[suf], location+filename[suf][ndx]);
     getmainin;
     new (indx);
     get_in (ass);
     get_in (rim);
     get_in (suf);
end; {procedure startup}

{END file loading routines}

function ordinal (qy : integer) : word;
{returns proper irregular ordinal multiplicative number from cardinal}
begin
   str (qy,sq);
   case qy of
        1   : ordinal := 'once';
        2   : ordinal := 'twice';
      else    ordinal := sq + ' times';
    end; {case}
end; {procedure ordinal}

function newword (x : entry) : word;
     {converts a record (back) to a line entry string}
var y : word;
    i,
    j : integer;
begin
     y := '                         ';
     with x do
       begin
          for i := 1 to length (spelling) do  {spelling is never empty}
            y[i] := spelling[i];
          j := spellength;
          if length (assonance) > 0 then  {assonance might be empty}
            for i := 1 to length (assonance) do
              y[j + i] := assonance[i];
          j := j + asslength;
          for i := 1 to length (rime) do  {rime is never empty}
           y[j + i] := rime[i];
          j := j + rimlength;
          if length (suffix) > 0 then   {suffix is probably empty}
            for i := 1 to length (suffix) do
              y[j + i] := suffix[i];
       end; {with statement}
     newword := y; {add a leading blank}
end; {function newword}

function Uwannaseeit (q : integer) : char;
{prompts for display and returns character}
var c : char;
begin
     write (' occurs '+ordinal (q)+' in the data.');
     if q = 1 then
        pron := 'it'
     else
        pron :='the list';
     gotoxy (16,8);
     queryloc;
     clreol;
     write ('Do you want to see '+pron+'? [Y]/N');
     read (kbd,c);
     if c in ['a'..'z'] then
        Uwannaseeit := upcase (c)
     else
        Uwannaseeit := c;
     write (c);
end; {char function Uwannaseeit}

function Randquest (q:integer) : char;
{prompts for random display and returns character}
var c : char;
begin
     at (lastloc);
     clreol;
     if firstrand then
       begin
          write ('Random equivalent set of ',q, '? Y/[N] ');
          firstrand := false;
       end
     else
       write ('Another random set of ',q,'? Y/[N] ');
     read (kbd,c);
     if c in ['a'..'z'] then
       Randquest := upcase (c)
     else
       Randquest := c;
     at (lastloc);
     clreol;
end; {char function Randquest}

procedure printout (q : integer);
{printer listing routine}
var xample : string [25];
    whosit : whatsit;
    topline : string[100];
begin
     writeln (lst);
     at (lastloc);
     clreol;
     write (' (Printing)');
     if itsa = rand then
      begin
        topline := '      Random equivalent of';
        whosit := lastone;
      end
     else
      begin
        topline := '               ';
        whosit := itsa;
      end;
     xample := example;
     if whosit = ass then
          xample := xample + ' -'
     else
        begin
          xample := '- ' + xample;
          if whosit = rim then
             topline := ' ' + topline;
          topline := '  ' + topline;
        end;
     case whosit of
       rim : topline := topline + ' Rime ';
       ass : topline := topline + ' Assonance ';
       suf : topline := topline + ' Suffix ';
      end; {case}
     topline := topline+xample+', which occurs '+ordinal (q)+' in the data';
     writeln (lst,topline);
     writeln (lst);
     j := 1;
     if odd (q) then
        offset := q div 2 + 1
     else
        offset := q div 2;
     while j <= offset do
       begin
          if   (j = offset)
           and (odd (q)) then
             writeln (lst, j:7,')   '+newword (ramfile[address[j]]^))
          else
             writeln (lst, j:7,')   '+newword (ramfile[address[j]]^),'  |  ',
                 (offset+j):7, ')   '+newword (ramfile[address[offset+j]]^));
          if j mod 10 = 0 then 
            write (^H'.)');
          if j mod 60 = 0 then
            begin {formfeed routine}
               writeln (lst, ^L);
               writeln (lst);
               writeln (^H'Page End)');
               write (' (New Page)');
               writeln (lst,topline);
               writeln (lst);
            end; {formfeed routine}
          j := j + 1;
       end; {while}
end; {procedure printout}

procedure filesave (q : integer);
{stores set requested in a disk file}
var xyz    : integer;
    rndext : string[4];
begin
     topline := example;
     for j := 1 to length (topline) do
       if topline[j] in ['a'..'z'] then
          topline[j] := upcase (topline[j]);
     case itsa  of
       rim : topline := topline + '.RIM';
       ass : topline := topline + '.ASS';
       suf : topline := topline + '.SUF';
      rand : begin
                str (random (1000):3, rndext);
                if rndext[1] = ' ' then
                  begin
                    rndext[1] := '0';
                    if rndext[2] = ' ' then
                       rndext[2] := '0';
                  end;
                topline := topline + '.' + rndext;
             end; {rand case}
      end; {case}
     at (lastloc);
     clreol;
     write (' (Writing file '+topline+')');
     assign (thisfile,topline);
     rewrite (thisfile);
     str (q,topline);
     topline := 'List of '+topline+' monosyllables ';
     if itsa = rand then
       begin
         topline := topline + 'selected at random for comparison ';
         topline := topline + 'with list of "'+example+'" words.';
       end
     else
       begin
          topline := topline + 'containing the ';
          case itsa of
            rim : topline := topline + 'rime -' + example + '.';
            suf : topline := topline + 'suffix -' + example + '.';
            ass : topline := topline + 'assonance ' + example + '-.';
           end; {case}
       end; {else}
     writeln (thisfile,topline);
     writeln (thisfile);
     for j := 1 to q do
       begin
          if j mod 10 = 0 then 
             write (^H'.)');
          writeln (thisfile, newword (ramfile[address[j]]^));
       end; {for}
     close (thisfile);
end; {procedure filesave}

procedure display (item_count : integer);
{handles screen display of set of items}
const
     first_margin = 5;
     second_margin = 11;
     main_margin = 23;
     first_pos = 22;
     main_pos = 40;
var
     j,
     margin,
     extra,
     increment,
     rimsz    : integer;
     one,
     topl     : string[15];
     whosit   : whatsit;
     topline1,
     topline2 : string [50];
     kays     : rime;
     orphan,
     single   : boolean;

     procedure topoff;  {internal to procedure display}
     {writes screen header}
     begin
          clrscr;
          if  single
            or   (orphan
             and (item_count - j <= 20)) then
              begin
                margin := main_margin;
                gotoxy (main_pos+extra,3);
                write (topl);
              end
          else
            begin
              margin := first_margin;
              gotoxy (first_pos+extra,3);
              write (topl);
              gotoxy (first_pos+extra+increment,3);
              write (topl);
            end;
          trademarked := false;
          gotoxy (9,1);
          reversv (topline1 + topline2);
          gotoxy (1,4);
     end; {internal procedure topoff}

begin {display proper}
     single := item_count <= 20;  {use single display for lists under 20 items}
     if single then
         increment := 20
     else
         increment := 40;
     if   (not single)        {if there's an orphan last screen}
      and (item_count mod increment <= 20) then 
        orphan := true
     else
        orphan := false;
     topl := temporary;
     if itsa = rand then
      begin
        topline1 := ' * R A N D O M *  ';
        topline2 := ' * E Q U I V A L E N T * ';
        whosit := lastone;
      end
     else
      begin
        topline1 := '                  ';
        topline2 := '                         ';
        whosit := itsa;
      end;
     case whosit of
          ass :  extra := 0;
          rim :  extra := asslength - 1;
          suf :  extra := rimlength + asslength - 1;
        end; {case}
     if item_count = 1 then
      begin
        sq := ' One';
        one := ' Word       ';
      end
     else
      begin
        str (item_count:4, sq);
        one := ' Words      ';
      end;
     topline2 := '       '+ sq + one + topline2;
     j := 1;
     topoff;
     while j <= item_count do
       begin
          write (j:margin, ')  ');
          reversv (' '+newword (ramfile[address[j]]^));
          if   (not single)
           and (j + 20 <= item_count) then
               begin
                  write (j+20:second_margin, ')  ');
                  reversv (' '+newword (ramfile[address[j+20]]^))
               end;
          writeln;
          if (j mod 20 = 0)
             or (j = item_count) then
             begin {newscreen}
                writeln;
                clreol;
                write ('                     ----> ');
                write ('Last, Next, or [Q]uit? ');
                if not single then    {readjust indices for double display}
                    if j + 20 > item_count then {if at the end}
                      j := item_count
                    else               {if there's more}
                      j := j + 20;
                cc := upcase (getkey);
                if escape = true then
                  begin
                    escape := false;
                    if  (cc  = 'K')
                     or (cc  = 'I')
                     or (cc  = 'H') then {if left or up arrow or PgUp}
                         cc := 'L';      {set to last}
                    if cc = 'G' then     {Home key -- restart at beginning}
                         j := 0;
                    if cc = 'Q' then     {PgDn key -- readjust to 'N'}
                         cc := 'N';
                    if cc = 'O' then     {End key -- display last screen}
                     begin
                       if item_count mod increment = 0 then
                         j := item_count - increment
                       else
                         j := (item_count div increment) * increment;
                       if orphan then
                          margin := main_margin;
                     end; {last screen display}
                  end;
                case cc of 'L' : begin {case of 'L'}
                        if j > (2 * increment) then
                           j := ((j-increment-1) div increment) * increment
                        else
                           j := 0;
                        if   orphan    {correct special orphan margin}
                         and (margin = main_margin) then
                           margin := first_margin;
                                 end; {case of 'L'}
                           'Q' : j := item_count;
                     end; {case cc}
                if j <> item_count then
                  topoff
                else
                  begin
                     write (^M);
                     clreol;
                     write ('                ----> ');
                     queryloc;
                     write ('Printout, FileSave, Both, or [Q]uit? ');
                     read (kbd,c);
                     if c in ['a'..'z'] then
                        c := upcase (c);
                  end; {if j = item_count}
             end; {newscreen..if single}
          j := j + 1;
       end; {while}
     case c of
      'B' : begin
              filesave (item_count);
              printout (item_count);
            end;
      'F' : filesave (item_count);
      'P' : printout (item_count);
        end; {case}
end; {procedure display}

procedure sorry;
{called if there are no hits}
begin
     writeln (' does not occur in the data.');
     clreol;
     writeln;
     clreol;
     write ('               -----> Press any key to continue ');
     read (kbd,c);
     write (^M);
     queryloc;
     clreol;
     lastloc[2] := lastloc[2] - 2;
     at (lastloc);
     clreol;
end; {procedure sorry}

procedure getemin (qq : integer);
{retrieves appropriate set of words}
var  topo,
     temp1,
     kk,
     t             : integer;
     norepeats     : boolean;
     dummy         : string [5];
begin
     case itsa of
   suf, rim : for j := 1 to qq do
                address[j] := indx^[itsa][adrs[itsa][k]+j] + 1;
        ass : for j := 1 to qq do
                address[j] := adrs[itsa][k]+j;
        rand : begin
                 randomize;
                 topo := number + 1;
                 address[1] := random (topo);
                 norepeats := true;
                 j := 2;
                 while j <= qq do
                    begin
                       temp1 := random (topo);
                       for kk := 1 to j-1 do
                           if address[kk] = temp1 then
                              norepeats := false;
                       if norepeats then
                         begin
                            address[j] := temp1;
                            j := j + 1;
                         end
                       else
                         norepeats := true;
                    end; {while}
          end; {case rand}
       end; {case statement}
     if itsa <> rand then
        gotoxy (16,10)
     else
        begin
           at (lastloc);
           lastloc[1] := lastloc[1] + 11;
        end;
end; {procedure getemin}

procedure getsuf;
{called from main menu to get instances of requested suffix}
var sufdone : boolean;
begin
  sufdone := false;
  lastone := suf;
  repeat
     itsa := suf;
     if not trademarked then
        trademark;
     for i := 4 to 9 do
       begin
          gotoxy (1,i);
          clreol;
       end;
     i := 0;
     gotoxy (16,4);
     write ('Please enter suffix to find (or <RET> to quit): ');
     queryloc;
     gotoxy (16,6);
     clreol;
     write ('(Enter <SPACE><RETURN> for Zero suffix)');
     at (lastloc);
     readln (testform);
     if testform = ' ' then
        testform := '   ';
     if testform = '' then
        sufdone := true
     else
        begin
           for i := 1 to length (testform) do
              if testform[i] in upper then
                 testform[i] := downcase (testform[i]);
           example := testform;
           if testform = '   ' then
              temporary := #$CD + '0'
           else
              temporary := #$CD + testform;
           writeln;
           clreol;
           write ('               The suffix ');
           reversv (' '+temporary+' ');
           while length (testform) < rimlength do
                testform := testform + ' ';
           found := false;
           for i := 1 to numof[suf] do
              if testform = the_list[suf][i]^ then
                 begin
                    found := true;
                    k := i;
                 end; {if and for}
           if not found then
             sorry
           else
             begin
              testform := temporary;
              firstrand := true;
              if Uwannaseeit (count[suf][k]) <> 'N' then
                 begin
                    getemin (count[suf][k]);
                    display (count[suf][k]);
                 end; {if U...}
              while Randquest (count[suf][k]) = 'Y' do
                 begin
                    itsa := rand;
                    getemin (count[suf][k]);
                    display (count[suf][k]);
                 end; {while Randquest..}
             end; { if not found else }
        end; {else}
   until sufdone;
end; {procedure getsuf}

procedure getass;
{called from main menu to get instances of requested assonance}
var assdone : boolean;
begin
   assdone := false;
   lastone := ass;
   repeat
     itsa := ass;
     i := 0;
     if not trademarked then
        trademark;
     for i := 4 to 9 do
       begin
          gotoxy (1,i);
          clreol;
       end;
     gotoxy (16,4);
     write ('Please enter assonance to find (or <RET> to quit): ');
     queryloc;
     gotoxy (16,6);
     clreol;
     write ('(Enter <SPACE><RETURN> for Zero assonance)');
     at (lastloc);
     readln (testform);
     if testform = ' ' then
        testform := '    ';
     if testform = '' then
       assdone := true
     else
       begin
          for i := 1 to length (testform) do
             if testform[i] in upper then
                testform[i] := downcase (testform[i]);
          writeln;
          example := testform;
          if testform = '    ' then
             temporary := '0' + #$CD
          else
             temporary := testform + #$CD;
          clreol;
          while length (testform) < rimlength do
             testform := testform + ' ';
          write ('               The assonance ');
          reversv (' '+temporary+' ');
          found := false;
          for i := 1 to numof[ass] do
             if testform = the_list[ass][i]^ then
               begin
                 found := true;
                 k := i;
               end;
          if not found then
            sorry
          else
            begin
               testform := temporary;
               firstrand := true;
               if Uwannaseeit (count[ass][k]) <> 'N' then
                 begin
                    getemin (count[ass][k]);
                    display (count[ass][k]);
                 end; {if Uwannaseeit}
               while Randquest (count[ass][k]) = 'Y' do
                 begin
                    itsa := rand;
                    getemin (count[ass][k]);
                    display (count[ass][k]);
                 end; {if Randquest}
            end; {else}
       end; {else}
   until assdone;
end; {procedure getass}

procedure getrime;
{called from main menu to get instances of requested rime}
var rimedone : boolean;
begin
   rimedone := false;
   lastone := rim;
   repeat
     itsa := rim;
     for i := 4 to 9 do
       begin
          gotoxy (1,i);
          clreol;
       end;
     i := 0;
     if not trademarked then
        trademark;
     gotoxy (16,4);
     write ('Please enter rime to find (or <RET> to quit): ');
     readln (testform);
     if testform = '' then
       rimedone := true
     else
       begin
          for i := 1 to length (testform) do
             if testform[i] in upper then
                testform[i] := downcase (testform[i]);
          example := testform;
          if testform = ' ' then
              temporary := #$CD + '0' + #$CD
          else
              temporary := #$CD + testform + #$CD;
          writeln;
          write ('               The rime ');
          reversv (' '+temporary+' ');
          while length (testform) < rimlength do
             testform := testform + ' ';
          found := false;
          for i := 1 to numof[rim] do
             if testform = the_list[rim][i]^ then
               begin
                 found := true;
                 k := i;
               end;
          if not found then
           sorry
          else
           begin
             firstrand := true;
             if Uwannaseeit (count[rim][k]) <> 'N' then
                begin
                  getemin (count[rim][k]);
                  display (count[rim][k]);
                end; {if Uwannaseeit}
             while Randquest (count[rim][k]) = 'Y' do
                begin
                  itsa := rand;
                  getemin (count[rim][k]);
                  display (count[rim][k]);
                end; {if Randquest}
           end; {if not found else}
       end; {else}
   until rimedone;
end; {procedure getrime}

begin {main program block}
     check_indices;
     startup;
     while not done do
       begin
          if not trademarked then
             trademark;
          for i := 4 to 9 do
            begin
               gotoxy (1,i);
               clreol;
            end;
          gotoxy (18,4);
          write ('Assonances, Rimes, Suffixes, [Q]uit? ');
          read (kbd,c);
          write (upcase(c));
          case c of
               'R', 'r'     : getrime;
               'a', 'A'     : getass;
               's', 'S'     : getsuf;
               'q', 'Q', ^M : done := true;
             end; {case}
          if not trademarked then
             trademark;
       end; {while not done}
     gotoxy (13,4);
     clreol;
     writeln ('                  Thank you.');
end. {main program block}
